@@ -1,8 +1,4 @@
-﻿"""Define the WikiText and SubWikiText classes."""
-
-# Todo: see if it is possible to use shadow position instead of using subpsans
-# Todo: consider using a tree structure (interval or segment tree).
-# Todo: Consider using separate strings for each node.
+"""Define the WikiText and SubWikiText classes."""
 
 from bisect import bisect_left, bisect_right, insort_right
 from copy import deepcopy
@@ -10,7 +6,8 @@ from html import unescape
 from itertools import islice
 from operator import attrgetter
 from typing import (
-    Dict, Generator, Iterable, List, MutableSequence, Optional, Tuple, Union)
+    Dict, Generator, Iterable, List, MutableSequence, Optional, Tuple, Type,
+    Union)
 from warnings import warn
 
 from regex import VERBOSE, DOTALL, MULTILINE, IGNORECASE, search, finditer
@@ -48,7 +45,7 @@ BRACKET_EXTERNAL_LINK = (
 EXTERNAL_LINK = \
     rb'(?>' + BARE_EXTERNAL_LINK + rb'|' + BRACKET_EXTERNAL_LINK + rb')'
 EXTERNAL_LINK_FINDITER = regex_compile(EXTERNAL_LINK, IGNORECASE).finditer
-INVALID_EXT_CHARS_SUB = regex_compile(  # the [:-4] slicee allows \[ and \]
+INVALID_EXT_CHARS_SUB = regex_compile(  # the [:-4] slice allows \[ and \]
     rb'[' + INVALID_URL_CHARS[:-4] + rb'{}]').sub
 
 # Sections
@@ -79,40 +76,31 @@ TABLE_FINDITER = regex_compile(
     """,
     DOTALL | MULTILINE | VERBOSE).finditer
 
-# Bolds
-BOLDS_FINDITER = regex_compile(
-    rb"""
-    (?>(
-        (?<=('\0*+'\0*+))'\0*+'\0*+'  # bold-italic start
-        |'\0*+'\0*+'
-    ))
-    # contents
-    \0*+[^'\n]++.*?
-    # bold end
-    (
-        '\0*+'\0*+'
-        (?=
-            (?(2)
-                (?:\0*+'\0*+')?+
-            )
-            (?>\0*+[^']|$)
-        )
-        |$
-    )
-    """,
+BOLD_ITALIC_FINDITER = regex_compile(  # bold-italic, bold, or italic tokens
+    rb"""((?>'\0*)*?)'\0*+'\0*+('\0*+('\0*+')?+)?+(?=[^']|$)|($)""",
     MULTILINE | VERBOSE).finditer
 
-ITALICS_FINDITER = regex_compile(
-    rb"""
+BOLD_FINDITER = regex_compile(rb"""
+    # start token
+    '\0*+'\0*+'
+    # content
+    (\0*+[^'\n]++.*?)
+    # end token
+    (?:'\0*+'\0*+'|$)
+""", MULTILINE | VERBOSE).finditer
+
+ITALIC_FINDITER = regex_compile(rb"""
+    # start token
     '\0*+'
-    # contents
-    \0*+[^'\n]++.*?
-    (?>
-        '\0*+'(?!\0*')
-        |()$
-    )
-    """,
-    MULTILINE | VERBOSE).finditer
+    # content
+    (\0*+[^'\n]++.*?)
+    # end token
+    (?:'\0*+'|$)
+""", MULTILINE | VERBOSE).finditer
+
+BOLD_ITALIC_RECURSE_METHODS = (
+    'templates', 'parser_functions', 'parameters', '_extension_tags',
+    'wikilinks')
 
 # Types which are detected by parse_to_spans
 SPAN_PARSER_TYPES = {
@@ -122,6 +110,32 @@ SPAN_PARSER_TYPES = {
 WS = '\r\n\t '
 
 
+class DeadIndexError(TypeError):
+    pass
+
+
+class DeadIndex(int):
+    """Do not allow adding to another integer but allow usage in a slice.
+
+    Addition of indices is the main operation during mutation of WikiText
+    objects.
+    """
+    __slots__ = ()
+
+    def __add__(self, o):
+        raise DeadIndexError(
+            'this usually means that the object has died '
+            '(overwritten or deleted) and cannot be mutated')
+    __radd__ = __add__
+
+    def __repr__(self):
+        return 'DeadIndex()'
+
+
+DEAD_INDEX = DeadIndex()  # == int() == 0
+DEAD_SPAN = DEAD_INDEX, DEAD_INDEX, None, None
+
+
 class WikiText:
 
     # In subclasses of WikiText _type is used as the key for _type_to_spans
@@ -393,7 +407,7 @@ class WikiText:
             ):
                 if e <= stop:
                     if ss != s or se != e:
-                        spans.pop(i + b)[:] = -1, -1, None, None
+                        spans.pop(i + b)[:] = DEAD_SPAN
                         b -= 1
 
     def _del_update(self, rmstart: int, rmstop: int) -> None:
@@ -434,7 +448,7 @@ class WikiText:
                         s, e, _, _ = span = spans[i]
                         continue
                     # rmstart <= s <= e < rmstop
-                    spans.pop(i)[:] = -1, -1, None, None
+                    spans.pop(i)[:] = DEAD_SPAN
                     i -= 1
                     if i < 0:
                         break
@@ -566,17 +580,15 @@ class WikiText:
         replace_external_links=True,
         replace_wikilinks=True,
         unescape_html_entities=True,
-        replace_bolds=True,
-        replace_italics=True,
+        replace_bolds_and_italics=True,
         _mutate=False
     ) -> str:
         """Return a plain text string representation of self."""
         if _mutate is False:
             s, e, m, b = self._span_data
-            parsed = WikiText(
-                [self._lststr[0][s:e]], self._inner_type_to_spans_copy())
-            parsed._span_data = [0, e - s, m, b[:]]
-            tts = parsed._type_to_spans
+            tts = self._inner_type_to_spans_copy()
+            parsed = WikiText([self._lststr[0][s:e]], tts)
+            parsed._span_data = tts[self._type][0]
         else:
             tts = self._type_to_spans
             parsed = self
@@ -595,11 +607,8 @@ class WikiText:
         # replacing bold and italics should be done before wikilinks and tags
         # because removing tags and wikilinks creates invalid spans, and
         # get_bolds() will try to look into wikilinks for bold parts.
-        if replace_bolds:
-            for b in parsed.get_bolds():
-                b[:] = b.text
-        if replace_italics:
-            for i in parsed.get_italics():
+        if replace_bolds_and_italics:
+            for i in reversed(parsed.get_bolds_and_italics()):
                 i[:] = i.text
         if replace_parameters:
             for p in parsed.parameters:
@@ -849,9 +858,143 @@ class WikiText:
             Comment(_lststr, _type_to_spans, span, 'Comment')
             for span in self._subspans('Comment')]
 
+    _relative_contents_end = span
+
     @property
-    def _relative_contents_end(self) -> int:
-        return self._span_data[1]
+    def _balanced_quotes_shadow(self):
+        """Return bold and italic match objects according MW's algorithm.
+
+        The comments at /includes/parser/Parser.php:doQuotes are helpful:
+        https://github.com/wikimedia/mediawiki/blob/master/includes/parser/Parser.php
+        https://phabricator.wikimedia.org/T15227#178834
+        """
+        bold_matches = []
+        odd_italics = False
+        odd_bold_italics = False
+        shadow_copy = self._shadow[:]
+        append_match = bold_matches.append
+        for match in BOLD_ITALIC_FINDITER(shadow_copy):
+            if match[4] is not None:  # newline or string end
+                if odd_italics is True and (
+                        len(bold_matches) + odd_bold_italics) % 2:
+                    # one of the bold marks needs to be interpreted as italic
+                    first_multi_letter_word = first_space = None
+                    for bold_match in bold_matches:
+                        bold_start = bold_match.start()
+                        if shadow_copy[bold_start - 1:bold_start] == b' ':
+                            if first_space is None:
+                                first_space = bold_start
+                            continue
+                        if shadow_copy[bold_start - 2:bold_start - 1] == b' ':
+                            shadow_copy[bold_start] = 95  # _
+                            break  # first_single_letter_word
+                        if first_multi_letter_word is None:
+                            first_multi_letter_word = bold_start
+                            continue
+                    else:  # there was no first_single_letter_word
+                        if first_multi_letter_word is not None:
+                            shadow_copy[first_multi_letter_word] = 95  # _
+                        elif first_space is not None:
+                            shadow_copy[first_space] = 95  # _
+                bold_matches.clear()
+                odd_italics = False
+                continue
+            if match[2] is None:  # italic
+                odd_italics ^= True
+                continue
+            if match[3] is None:  # bold
+                s, e = match.span(1)
+                if s != e:  # four apostrophes, hide the first one
+                    shadow_copy[s] = 95  # _
+                append_match(match)
+                continue
+            # bold-italic
+            s, e = match.span(1)
+            es = e - s
+            if es:  # more than 5 apostrophes, hide the previous ones
+                shadow_copy[s:e] = b'_' * es
+            odd_bold_italics ^= True
+            odd_italics ^= True
+        return shadow_copy
+
+    def get_bolds_and_italics(
+        self, *, recursive=True, filter_cls: type = None
+    ) -> List[Union['Bold', 'Italic']]:
+        """Return a list of bold and italic objects in self.
+
+        This is faster than calling ``get_bolds`` and ``get_italics``
+        individually.
+        :keyword recursive: if True also look inside templates, parser
+            functions, extension tags, etc.
+        :keyword filter_cls: only return this type. Should be
+            `wikitextparser.Bold` or `wikitextparser.Italic`.
+            The default is None and means both bolds and italics.
+        """
+        result = []
+        append = result.append
+        _lststr = self._lststr
+        s = self._span_data[0]
+        type_to_spans = self._type_to_spans
+        tts_setdefault = type_to_spans.setdefault
+        balanced_shadow = self._balanced_quotes_shadow
+        rs, re = self._relative_contents_end
+
+        if filter_cls is None or filter_cls is Bold:
+            bold_spans = tts_setdefault('Bold', [])
+            get_old_bold_span = {(s[0], s[1]): s for s in bold_spans}.get
+            bold_matches = list(BOLD_FINDITER(balanced_shadow, rs, re))
+            for match in bold_matches:
+                ms, me = match.span()
+                b, e = s + ms, s + me
+                old_span = get_old_bold_span((b, e))
+                if old_span is None:
+                    span = [b, e, None, balanced_shadow[ms:me]]
+                    insort_right(bold_spans, span)
+                else:
+                    span = old_span
+                append(Bold(_lststr, type_to_spans, span, 'Bold'))
+            if not recursive:
+                return result
+            for m in BOLD_ITALIC_RECURSE_METHODS:
+                for e in getattr(self, m):
+                    result += e.get_bolds(False)
+            if filter_cls is Bold:
+                return result
+        else:  # filter_cls is Italic
+            bold_matches = BOLD_FINDITER(balanced_shadow, rs, re)
+
+        # filter_cls is None or filter_cls is Italic
+
+        # remove bold tokens before searching for italics
+        for match in bold_matches:
+            ms, me = match.span()
+            cs, ce = match.span(1)  # content
+            balanced_shadow[ms:cs] = b'_' * (cs - ms)
+            balanced_shadow[ce:me] = b'_' * (me - ce)
+
+        italic_spans = tts_setdefault('Italic', [])
+        get_old_italic_span = {(s[0], s[1]): s for s in italic_spans}.get
+        for match in ITALIC_FINDITER(balanced_shadow, rs, re):
+            ms, me = match.span()
+            b, e = span = s + ms, s + me
+            old_span = get_old_italic_span(span)
+            if old_span is None:
+                span = [b, e, None, balanced_shadow[ms:me]]
+                insort_right(italic_spans, span)
+            else:
+                span = old_span
+            append(Italic(
+                _lststr, type_to_spans, span, 'Bold', me != match.end(1)))
+        if not recursive:
+            result.sort(key=attrgetter('_span_data'))
+            return result
+        for m in BOLD_ITALIC_RECURSE_METHODS:
+            for e in getattr(self, m):
+                result += e.get_italics(False)
+
+        if filter_cls is None:  # the results are mixed and unsorted
+            result.sort(key=attrgetter('_span_data'))
+        return result
 
     def get_bolds(self, recursive=True) -> List['Bold']:
         """Return bold parts of self.
@@ -859,36 +1002,7 @@ class WikiText:
         :param recursive: if True also look inside templates, parser functions,
             extension tags, etc.
         """
-        _lststr = self._lststr
-        type_to_spans = self._type_to_spans
-        s = self._span_data[0]
-        spans = type_to_spans.setdefault('Bold', [])
-        span_tuple_to_span_get = {(s[0], s[1]): s for s in spans}.get
-        bolds = []
-        bolds_append = bolds.append
-        shadow = self._shadow
-        for match in BOLDS_FINDITER(
-            shadow, endpos=self._relative_contents_end
-        ):
-            ms = match.start(1)
-            me = match.end(3)
-            b, e = s + ms, s + me
-            old_span = span_tuple_to_span_get((b, e))
-            if old_span is None:
-                span = [b, e, None, shadow[ms:me]]
-                insort_right(spans, span)
-            else:
-                span = old_span
-            bolds_append(Bold(_lststr, type_to_spans, span, 'Bold'))
-        if not recursive:
-            return bolds
-        for t in (
-            'templates', 'parser_functions', 'parameters', '_extension_tags',
-            'wikilinks'
-        ):
-            for e in getattr(self, t):
-                bolds += e.get_bolds(False)
-        return bolds
+        return self.get_bolds_and_italics(filter_cls=Bold, recursive=recursive)
 
     def get_italics(self, recursive=True) -> List['Italic']:
         """Return italic parts of self.
@@ -896,42 +1010,8 @@ class WikiText:
         :param recursive: if True also look inside templates, parser functions,
             extension tags, etc.
         """
-        shadow_copy = self._shadow[:]
-        # remove bolds
-        for match in BOLDS_FINDITER(shadow_copy):
-            s, e = match.span(1)
-            shadow_copy[s:e] = b'B' * (e - s)
-            s, e = match.span(3)
-            shadow_copy[s:e] = b'B' * (e - s)
-        type_to_spans = self._type_to_spans
-        s = self._span_data[0]
-        _lststr = self._lststr
-        spans = type_to_spans.setdefault('Italic', [])
-        span_tuple_to_span_get = {(s[0], s[1]): s for s in spans}.get
-        italics = []
-        italics_append = italics.append
-        for match in ITALICS_FINDITER(
-            shadow_copy, endpos=self._relative_contents_end
-        ):
-            ms, me = match.span()
-            b, e = span = s + ms, s + me
-            old_span = span_tuple_to_span_get(span)
-            if old_span is None:
-                span = [b, e, None, shadow_copy[ms:me]]
-                insort_right(spans, span)
-            else:
-                span = old_span
-            italics_append(Italic(
-                _lststr, type_to_spans, span, 'Bold', match[1] is None))
-        if not recursive:
-            return italics
-        for t in (
-            'templates', 'parser_functions', 'parameters', '_extension_tags',
-            'wikilinks'
-        ):
-            for e in getattr(self, t):
-                italics += e.get_italics(False)
-        return italics
+        return self.get_bolds_and_italics(
+            filter_cls=Italic, recursive=recursive)
 
     @property
     def external_links(self) -> List['ExternalLink']:
